@@ -1,16 +1,19 @@
 """A friendly Python SFTP interface."""
+from __future__ import print_function
 
 import os
+from contextlib import contextmanager
 import socket
 from stat import S_IMODE, S_ISDIR, S_ISREG
 import tempfile
 
 import paramiko
-from paramiko import SSHException   # make available
+from paramiko import SSHException              # make available
 from paramiko import AuthenticationException   # make available
 from paramiko import AgentKey
 
-__version__ = "0.2.7"
+__version__ = "0.2.8"
+# pylint: disable = R0913
 
 
 def st_mode_to_int(val):
@@ -24,7 +27,7 @@ def st_mode_to_int(val):
     :returns int: integer representation of octal mode
 
     '''
-    return int(str(oct(S_IMODE(val))))
+    return int(str(oct(S_IMODE(val)))[-3:])
 
 
 class ConnectionException(Exception):
@@ -65,21 +68,21 @@ class WTCallbacks(object):
         self.ulist = []
 
     def file_cb(self, pathname):
-        '''called for regular files
+        '''called for regular files, appends pathname to .flist
 
         :param str pathname: file path
         '''
         self.flist.append(pathname)
 
     def dir_cb(self, pathname):
-        '''called for directories
+        '''called for directories, appends pathname to .dlist
 
         :param str pathname: directory path
         '''
         self.dlist.append(pathname)
 
     def unk_cb(self, pathname):
-        '''called for unknown file types
+        '''called for unknown file types, appends pathname to .ulist
 
         :param str pathname: unknown entity path
         '''
@@ -90,22 +93,31 @@ class Connection(object):
     """Connects and logs into the specified hostname.
     Arguments that are not given are guessed from the environment.
 
-    :param str host: The Hostname or IP of the remote machine.
-    :param str username: Your username at the remote machine.
-    :param private_key: path to private key file(str) or paramiko.AgentKey
-    :param str password: Your password at the remote machine.
-    :param int port: The SSH port of the remote machine.(default: 22)
-    :param str private_key_pass: password to use, if private_key is encrypted.
-    :param list ciphers: List of ciphers to use in order.
-    :param bool|str log:
-        log connection/handshake details? (default=False) if set to True,
+    :param str host:
+        The Hostname or IP of the remote machine.
+    :param str|None username: *Default: None* -
+        Your username at the remote machine.
+    :param str|obj|None private_key: *Default: None* -
+        path to private key file(str) or paramiko.AgentKey
+    :param str|None password: *Default: None* -
+        Your password at the remote machine.
+    :param int port: *Default: 22* -
+        The SSH port of the remote machine.
+    :param str|None private_key_pass: *Default: None* -
+        password to use, if private_key is encrypted.
+    :param list|None ciphers: *Default: None* -
+        List of ciphers to use in order.
+    :param bool|str log: *Default: False* -
+        log connection/handshake details? If set to True,
         pysftp creates a temporary file and logs to that.  If set to a valid
         path and filename, pysftp logs to that.  The name of the logfile can
         be found at  ``.logfile``
     :returns: (obj) connection to the requested host
-    :raises:
-        ConnectionException, CredentialException, SSHException,
-        AuthenticationException, PasswordRequiredException
+    :raises ConnectionException:
+    :raises CredentialException:
+    :raises SSHException:
+    :raises AuthenticationException:
+    :raises PasswordRequiredException:
 
     """
 
@@ -227,30 +239,28 @@ class Connection(object):
 
         :param str remotedir: the remote directory to copy from (source)
         :param str localdir: the local directory to copy to (target)
-        :param bool preserve_mtime:
-            preserve modification time on files(default: False)
+        :param bool preserve_mtime: *Default: False* -
+            preserve modification time on files
 
         :returns: None
 
         :raises:
         """
         self._sftp_connect()
-        saved = self.pwd
-        self.cwd(remotedir)
-        for sattr in self._sftp.listdir_attr('.'):
-            if S_ISREG(sattr.st_mode):
-                rname = sattr.filename
-                self.get(rname, reparent(localdir, rname),
-                         preserve_mtime=preserve_mtime)
-        self.cwd(saved)
+        with self.cd(remotedir):
+            for sattr in self._sftp.listdir_attr('.'):
+                if S_ISREG(sattr.st_mode):
+                    rname = sattr.filename
+                    self.get(rname, reparent(localdir, rname),
+                             preserve_mtime=preserve_mtime)
 
     def get_r(self, remotedir, localdir, preserve_mtime=False):
         """recursively copy remotedir structure to localdir
 
         :param str remotedir: the remote directory to copy from
         :param str localdir: the local directory to copy to
-        :param bool preserve_mtime:
-            preserve modification time on files(default: False)
+        :param bool preserve_mtime: *Default: False* -
+            preserve modification time on files
 
         :returns: None
 
@@ -323,7 +333,8 @@ class Connection(object):
         :returns:
             (obj) SFTPAttributes containing attributes about the given file
 
-        :raises: IOError, OSError
+        :raises IOError: if remotepath doesn't exist
+        :raises OSError: if localpath doesn't exist
 
         """
         if not remotepath:
@@ -342,6 +353,85 @@ class Connection(object):
 
         return sftpattrs
 
+    def put_d(self, localpath, remotepath, confirm=True, preserve_mtime=False):
+        """Copies a local directory's contents to a remotepath
+
+        :param str localpath: the local path to copy (source)
+        :param str remotepath:
+            the remote path to copy to (target)
+        :param bool confirm:
+            whether to do a stat() on the file afterwards to confirm the file
+            size
+        :param bool preserve_mtime:
+            *Default: False* - make the modification time(st_mtime) on the
+            remote file match the time on the local. (st_atime can differ
+            because stat'ing the localfile can/does update it's st_atime)
+
+        :returns: None
+
+        :raises IOError: if remotepath doesn't exist
+        :raises OSError: if localpath doesn't exist
+        """
+        self._sftp_connect()
+        wtcb = WTCallbacks()
+        cur_local_dir = os.getcwd()
+        os.chdir(localpath)
+        walktree('.', wtcb.file_cb, wtcb.dir_cb, wtcb.unk_cb,
+                 recurse=False)
+        for fname in wtcb.flist:
+            src = os.path.join(localpath, fname)
+            dest = reparent(remotepath, fname)
+            # print('put', src, dest)
+            self.put(src, dest, confirm=confirm, preserve_mtime=preserve_mtime)
+
+        # restore local directory
+        os.chdir(cur_local_dir)
+
+    def put_r(self, localpath, remotepath, confirm=True, preserve_mtime=False):
+        """Recursively copies a local directory's contents to a remotepath
+
+        :param str localpath: the local path to copy (source)
+        :param str remotepath:
+            the remote path to copy to (target)
+        :param bool confirm:
+            whether to do a stat() on the file afterwards to confirm the file
+            size
+        :param bool preserve_mtime:
+            *Default: False* - make the modification time(st_mtime) on the
+            remote file match the time on the local. (st_atime can differ
+            because stat'ing the localfile can/does update it's st_atime)
+
+        :returns: None
+
+        :raises IOError: if remotepath doesn't exist
+        :raises OSError: if localpath doesn't exist
+        """
+        self._sftp_connect()
+        wtcb = WTCallbacks()
+        cur_local_dir = os.getcwd()
+        os.chdir(localpath)
+        walktree('.', wtcb.file_cb, wtcb.dir_cb, wtcb.unk_cb)
+        # restore local directory
+        os.chdir(cur_local_dir)
+        for dname in wtcb.dlist:
+            #for subdir in path_advance(dname):
+            if dname != '.':
+                self.mkdir(reparent(remotepath, dname))
+
+        for fname in wtcb.flist:
+            head, _ = os.path.split(fname)
+            if head not in wtcb.dlist:
+                for subdir in path_advance(head):
+                    if subdir not in wtcb.dlist and subdir != '.':
+                        self.mkdir(reparent(remotepath, subdir))
+                        wtcb.dlist.append(subdir)
+            src = os.path.join(localpath, fname)
+            dest = reparent(remotepath, fname)
+            # print('put', src, dest)
+            self.put(src, dest, confirm=confirm, preserve_mtime=preserve_mtime)
+
+
+
     def putfo(self, flo, remotepath=None, file_size=0, callback=None,
               confirm=True):
 
@@ -388,6 +478,24 @@ class Connection(object):
         else:
             return channel.makefile_stderr('rb', -1).readlines()
 
+    @contextmanager
+    def cd(self, remotepath=None):
+        """context manager that can change to a optionally specified remote
+        directory and restores the old pwd on exit.
+
+        :param str|None remotepath: *Default: None* -
+            remotepath to temporarily make the current directory
+        :returns: None
+        :raises: IOError, if remote path doesn't exist
+        """
+        try:
+            original_path = self.pwd
+            if remotepath is not None:
+                self.cwd(remotepath)
+            yield
+        finally:
+            self.cwd(original_path)
+
     def chdir(self, remotepath):
         """change the current working directory on the remote
 
@@ -408,8 +516,8 @@ class Connection(object):
         representation of the octal mode to use.
 
         :param str remotepath: the remote path/file to modify
-        :param int mode:
-            int representation of octal mode for directory, default 777
+        :param int mode: *Default: 777* -
+            int representation of octal mode for directory
 
         :returns: None
 
@@ -461,7 +569,7 @@ class Connection(object):
 
         :param str remotepath: path to list on the server
 
-        :returns: (list) directory entries, sorted
+        :returns: (list of str) directory entries, sorted
 
         """
         self._sftp_connect()
@@ -479,11 +587,12 @@ class Connection(object):
 
         :param str remotepath: path to list on the server
 
-        :returns: (list of SFTPAttributes)
+        :returns: (list of SFTPAttributes), sorted
 
         """
         self._sftp_connect()
-        return self._sftp.listdir_attr(remotepath)
+        return sorted(self._sftp.listdir_attr(remotepath),
+                      key=lambda attr: attr.filename)
 
     def mkdir(self, remotepath, mode=777):
         """Create a directory named remotepath with mode. On some systems,
@@ -491,8 +600,8 @@ class Connection(object):
         masked out.
 
         :param str remotepath: directory to create`
-        :param int mode:
-            int representation of octal mode for directory, default 777
+        :param int mode: *Default: 777* -
+            int representation of octal mode for directory
 
         :returns: None
 
@@ -552,8 +661,8 @@ class Connection(object):
         in the way, raise an exception.
 
         :param str remotedir: the directory structure to create
-        :param int mode:
-            int representation of octal mode for directory, default 777
+        :param int mode: *Default: 777* -
+            int representation of octal mode for directory
 
         :returns: None
 
@@ -672,7 +781,7 @@ class Connection(object):
         :param str remote_file: name of the file to open.
         :param str mode:
             mode (Python-style) to open file (always assumed binary)
-        :param int bufsize: desired buffering (-1 = default buffer size)
+        :param int bufsize: *Default: -1* - desired buffering
 
         :returns: (obj) SFTPFile, a handle the remote open file
 
@@ -746,7 +855,7 @@ class Connection(object):
         self._sftp.truncate(remotepath, size)
         return self._sftp.stat(remotepath).st_size
 
-    def walktree(self, remotepath, fcallback, dcallback, ucallback):
+    def walktree(self, remotepath, fcallback, dcallback, ucallback, recurse=True):
         '''recursively descend, depth first, the directory tree rooted at
         remotepath, calling discreet callback functions for each regular file,
         directory and unknown file type.
@@ -762,6 +871,7 @@ class Connection(object):
         :param callable ucallback:
             callback function to invoke for an unknown file type.
             (form: ``func(str)``)
+        :param bool recurse: *Default: True* - should it recurse
 
         :returns: None
 
@@ -775,8 +885,9 @@ class Connection(object):
             if S_ISDIR(mode):
                 # It's a directory, call the dcallback function
                 dcallback(pathname)
-                # now, recurse into it
-                self.walktree(pathname, fcallback, dcallback, ucallback)
+                if recurse:
+                    # now, recurse into it
+                    self.walktree(pathname, fcallback, dcallback, ucallback)
             elif S_ISREG(mode):
                 # It's a file, call the fcallback function
                 fcallback(pathname)
@@ -832,8 +943,9 @@ class Connection(object):
 
     @property
     def timeout(self):
-        ''' (float|None) get or set the underlying socket timeout for pending
-        read/write ops.  (default: None - no timeout)
+        ''' (float|None) *Default: None* -
+            get or set the underlying socket timeout for pending read/write
+            ops.
 
         :returns:
             (float|None) seconds to wait for a pending read/write operation
@@ -866,7 +978,7 @@ def path_advance(thepath, sep=os.sep):
     '''generator to iterate over a file path forwards
 
     :param str thepath: the path to navigate forwards
-    :param str sep: the path separator to use, defaults to ``os.sep``
+    :param str sep: *Default: os.sep* - the path separator to use
 
     :returns: (iter)able of strings
 
@@ -892,7 +1004,7 @@ def path_retreat(thepath, sep=os.sep):
     '''generator to iterate over a file path in reverse
 
     :param str thepath: the path to retreat over
-    :param str sep: the path separator to use, default to ``os.sep``
+    :param str sep: *Default: os.sep* - the path separator to use
 
     :returns: (iter)able of strings
 
@@ -922,3 +1034,62 @@ def reparent(newparent, oldpath):
     if oldpath[0] == os.sep:
         oldpath = '.' + oldpath
     return os.path.join(newparent, oldpath)
+
+
+def walktree(localpath, fcallback, dcallback, ucallback, recurse=True):
+    '''on the local file system, recursively descend, depth first, the
+    directory tree rooted at localpath, calling discreet callback functions
+    for each regular file, directory and unknown file type.
+
+    :param str localpath:
+        root of remote directory to descend, use '.' to start at
+        :attr:`.pwd`
+    :param callable fcallback:
+        callback function to invoke for a regular file.
+        (form: ``func(str)``)
+    :param callable dcallback:
+        callback function to invoke for a directory. (form: ``func(str)``)
+    :param callable ucallback:
+        callback function to invoke for an unknown file type.
+        (form: ``func(str)``)
+    :param bool recurse: *Default: True* -  should it recurse
+
+    :returns: None
+
+    :raises: OSError, if localpath doesn't exist
+
+    '''
+
+    for entry in os.listdir(localpath):
+        pathname = os.path.join(localpath, entry)
+        mode = os.stat(pathname).st_mode
+        if S_ISDIR(mode):
+            # It's a directory, call the dcallback function
+            dcallback(pathname)
+            if recurse:
+                # now, recurse into it
+                walktree(pathname, fcallback, dcallback, ucallback)
+        elif S_ISREG(mode):
+            # It's a file, call the fcallback function
+            fcallback(pathname)
+        else:
+            # Unknown file type
+            ucallback(pathname)
+
+@contextmanager
+def cd(localpath=None):
+    """context manager that can change to a optionally specified local
+    directory and restores the old pwd on exit.
+
+    :param str|None localpath: *Default: None* -
+        local path to temporarily make the current directory
+    :returns: None
+    :raises: OSError, if local path doesn't exist
+    """
+    try:
+        original_path = os.getcwd()
+        if localpath is not None:
+            os.chdir(localpath)
+        yield
+    finally:
+        os.chdir(original_path)
